#include "client/Game.h"
#include "client/GameClient.h"
#include "client/packets/WorldPackets.h"
#include "client/rendering/Engine.h"
#include "common/network/Packets.h"
#include "utils/Logger.h"
#include "utils/Utils.h"

Game* Game::game = nullptr;
BlockRegistry Game::blockRegistry;
World Game::world{blockRegistry};

Game::Game(const Controller& controller)
    : controller(controller), tickState(&Game::tickConnectState),
      renderState(&Game::renderConnectState), baseGUI(controller),
      startGUI(baseGUI), worldRenderer(world) {
    pos = Vector3(0.0f, 30.0f, 0.0f);
    rotation = Quaternion(Vector3(1.0f, 0.0f, 0.0f), 30) * rotation;
    rotation = Quaternion(Vector3(0.0f, 1.0f, 0.0f), 30) * rotation;
    game = this;
}

void Game::tick() {
    (this->*tickState)();

    lastRotation = rotation;
    lastPos = pos;

    Vector3 right = rotation * Vector3(1.0f, 0.0f, 0.0f);
    Vector3 up = rotation * Vector3(0.0f, 1.0f, 0.0f);
    Vector3 back = rotation * Vector3(0.0f, 0.0f, -1.0f);

    const float speed = 2.0f;
    if(controller.down.isDown()) {
        pos += back * speed;
    }
    if(controller.up.isDown()) {
        pos -= back * speed;
    }
    if(controller.left.isDown()) {
        pos -= right * speed;
    }
    if(controller.right.isDown()) {
        pos += right * speed;
    }
    if(controller.jump.isDown()) {
        pos += up * speed;
    }
    if(controller.sneak.isDown()) {
        pos -= up * speed;
    }

    const float rotationSpeed = 5.0f;
    if(controller.camLeft.isDown()) {
        rotation = Quaternion(up, -rotationSpeed) * rotation;
    }
    if(controller.camRight.isDown()) {
        rotation = Quaternion(up, rotationSpeed) * rotation;
    }
    if(controller.camUp.isDown()) {
        rotation = Quaternion(right, -rotationSpeed) * rotation;
    }
    if(controller.camDown.isDown()) {
        rotation = Quaternion(right, rotationSpeed) * rotation;
    }
}

void Game::renderWorld(float lag, ShaderMatrix& sm) {
    sm.update(Utils::interpolate(lastPos, pos, lag),
              lastRotation.lerp(lag, rotation));
    worldRenderer.render(lag, sm);
}

void Game::renderOverlay(float lag, ShaderMatrix& sm, Renderer& r) {
    (this->*renderState)(lag, sm, r);
    sm.identity().scale(2.0f).update();
    StringBuffer<100> s;
    s.append("FPS: &074")
        .append(Engine::getFrameClock().getUpdatesPerSecond())
        .append(" &999TPS: &722")
        .append(Engine::getTickClock().getUpdatesPerSecond());
    r.renderString(Vector2(10.0f, 10.0f), s);
}

bool Game::isRunning() const {
    return true;
}

void Game::tickConnectState() {
    startGUI.tick();
    StartGUI::Address a;
    if(startGUI.getAddress(a)) {
        Error error = GameClient::connect(a, 11196, 3000);
        if(error.has()) {
            LOG_INFO(error.message);
        } else {
            LOG_INFO("connected");
            tickState = &Game::tickConnectedState;
            renderState = &Game::renderConnectedState;
        }
    }
}

void Game::tickConnectedState() {
    GameClient::consumeEvents();
}

void Game::renderConnectState(float lag, ShaderMatrix& sm, Renderer& r) {
    startGUI.render(lag, sm, r);
}

void Game::renderConnectedState(float lag, ShaderMatrix& sm, Renderer& r) {
    (void)lag;
    (void)sm;
    (void)r;
}